from .common_webscraper_functions import fetch_soup_from_page, row_has_link, get_tr_of_stats_table, get_tr_of_table_with_id, true
from .field_parser import FieldParserFactory
from .object_fetcher import ObjectFetcherFactory, HTMLObjectIterator, HTMLObjectIteratorFactory
import logging

class CreateFromPageParserFactory:

    def __init__(self, create_from_page_parser_dict) -> None:
        self.create_from_page_parser = CreateFromPageParser(create_from_page_parser_dict['base_url'], DataDictParserFactory(create_from_page_parser_dict['parser']).data_dict_parser)


class CreateFromPageParser:

    def __init__(self, base_url, data_dict_parser) -> None:
        self.base_url = base_url
        self.data_dict_parser = data_dict_parser

    def parse(self, url, webscraperObjectCollection):
        soup = fetch_soup_from_page(self.base_url + url)
        data_dict = self.data_dict_parser.parse(soup, {}, webscraperObjectCollection)
        data_dict['url'] = url
        return data_dict

class TableParserObject:

    def __init__(self, all_object_selection_function, narrow_down_function, data_dict_parser):
        self.all_object_selection_function = all_object_selection_function
        self.narrow_down_function = narrow_down_function
        self.data_dict_parser = data_dict_parser

    def parse_page(self, soup, data_dict, webscraperObjectCollection):
        return_array = []
        for eligible_element in self.all_object_selection_function(soup):
            if self.narrow_down_function(eligible_element):
                return_array.append(self.data_dict_parser.parse(eligible_element, data_dict, webscraperObjectCollection))
        return return_array

class GenericParserObject:

    def __init__(self, html_object_iterator: HTMLObjectIterator, data_dict_parser):
        self.html_object_iterator = html_object_iterator
        self.data_dict_parser = data_dict_parser

    def parse_page(self, soup, data_dict, webscraper_object_collection):
        return_array = []
        for eligible_element in self.html_object_iterator.get_valid_elements(soup):
            print(eligible_element)
            return_array.append(self.data_dict_parser.parse(eligible_element, data_dict, webscraper_object_collection))
        return return_array 

class ParserObject:

    def __init__(self, base_object_function, children_element_function):
        self.base_object_function = base_object_function
        self.children_element_function = children_element_function

    def parse_page(self, soup, data_dict, webscraperObjectCollection):
        return_array = []
        for eligible_element in self.children_element_function(self.base_object_function(soup)):
            return_array.append(self.data_dict_parser.parse(eligible_element, data_dict, webscraperObjectCollection))
        return return_array

class ParserObjectFactory:

    def _get_narrow_down_function(self, function_name):
        if function_name == 'row_has_link':
            return row_has_link
        elif function_name == 'true':
            return true
        else:
            raise Exception("No match for narrow down function")


    def __init__(self, parser_dict):
        self.parser_dict = parser_dict
        if 'html_object_iterator' in parser_dict.keys():
            html_object_iterator = HTMLObjectIteratorFactory(parser_dict['html_object_iterator']).create()
            data_dict_parser = DataDictParserFactory(parser_dict['data_dict_parser']).data_dict_parser
            self.parser = GenericParserObject(html_object_iterator, data_dict_parser)
        elif parser_dict['parser_type'] == 'table':
            if 'table_id' in parser_dict.keys():
                self.parser = TableParserObject(get_tr_of_table_with_id(parser_dict['table_id']), self._get_narrow_down_function(parser_dict['narrow_down_function']), DataDictParserFactory(parser_dict['data_dict_parser']).data_dict_parser)
            else:
                self.parser = TableParserObject(get_tr_of_stats_table(), self._get_narrow_down_function(parser_dict['narrow_down_function']), DataDictParserFactory(parser_dict['data_dict_parser']).data_dict_parser)
        elif parser_dict['parser_type'] == 'generic':
            self.parser = ParserObject(get_element(parser_dict['base_object']), get_children_element(parser_dict['base_element']))
        else:
            raise Exception("No match for parser type")


class DataDictParserFactory:

    def __init__(self, data_dict_parser_dict):
        logging.info('[DataDictParserFactory] [Init] ' + str(data_dict_parser_dict))
        field_parsers = []
        for field_dict in data_dict_parser_dict:
            field_parsers.append(FieldParserFactory(field_dict).get_field_parser())
        self.data_dict_parser = DataDictParser(field_parsers)

class DataDictParser:

    def __init__(self, field_parsers):
        self.field_parsers = field_parsers

    def parse(self, html_object, data_dict, webscraperObject):
        return_dict = {}
        for field_parser in self.field_parsers:
            return_dict[field_parser.field_name] = field_parser.parse(html_object, data_dict, webscraperObject)
        return return_dict


from bradleystevenson2015_database import database
import json
from bradleystevenson2015_webscraper.common_webscraper_functions import fetch_soup_from_page
from bradleystevenson2015_webscraper.parser import CreateFromPageParserFactory, ParserObjectFactory
from bradleystevenson2015_webscraper.url_generator import URLGeneratorFactory
import logging


class WebscraperObjectCollection:

    def __init__(self, webscraper_schema_filepath, database_path, database_schema_filepath, custom_objects):
        self.databaseObject = database.Database(database_path, database_schema_filepath)
        self._create_webscraper_objects(webscraper_schema_filepath, custom_objects)


    def _create_webscraper_objects(self, webscraper_schema_filepath, custom_objects):
        self.webscrapers = []
        file = open(webscraper_schema_filepath)
        data = json.load(file)
        file.close()
        for webscraper_object in data['objects']:
            self.webscrapers.append(WebscraperObjectFactory(webscraper_object, custom_objects).webscraper)


    def get_webscraper_object_with_name(self, object_name):
        for webscraper in self.webscrapers:
            if webscraper.object_name == object_name:
                return webscraper

    def run(self, arguments):
        if '--create-tables' in arguments:
            self.databaseObject.create_tables()
            exit(0)
        create_from_web_dict = self._parse_arguments(arguments)
        for webscraper in self.webscrapers:
            webscraper.create(create_from_web_dict[webscraper.object_name], self)
        self.databaseObject.insert_into_database()

    def _parse_arguments(self, arguments):
        logging.info("[WEBSCRAPER] [_PARSE_ARGUMENTS] " + str(arguments))
        return_dict = {}
        for webscraper in self.webscrapers:
            return_dict[webscraper.object_name] = False
        for argument in arguments[1:]:
            if argument not in return_dict.keys() and argument != 'all':
                raise Exception("No match for object name")
            return_dict[argument] = True
        if 'all' in arguments[1:]:
            for key in return_dict.keys():
                return_dict[key] = True
        return return_dict


class WebscraperObject:

    def __init__(self, object_name, tables, create_from_page_parser=None):
        self.object_name = object_name
        self.tables = tables
        self.create_from_page_parser = create_from_page_parser

    def create(self, create_from_web, webscraperObjectCollection):
        logging.info("[WebscraperObject] [create] " + self.object_name + " " + str(create_from_web))
        if create_from_web:
            self.create_from_web(webscraperObjectCollection)
        else:
            self.create_from_database(webscraperObjectCollection)

    def create_from_web(self, webscraperObjectCollection):
        pass

    def create_from_database(self, webscraperObjectCollection):
        for table_name in self.tables:
            webscraperObjectCollection.databaseObject.tables[table_name].generate_from_database()

    def create_from_page(self, url, webscraperObjectCollection):
        if self.create_from_page_parser is None:
            raise Exception("We have no way to create this object")
        data_dict = self.create_from_page_parser.parse(url, webscraperObjectCollection)
        data_dict['url'] = url
        return webscraperObjectCollection.databaseObject.tables[self.tables[0]].append(data_dict)
    
class NewWebscraperObject(WebscraperObject):

    def __init__(self, object_name, parsers, url_generator, create_from_page_parser):
        self.object_name = object_name
        self.parsers = parsers
        self.create_from_page_parser = create_from_page_parser
        self.url_generator = url_generator
        super().__init__(object_name, [object_name], create_from_page_parser)

    def create_from_web(self, webscraperObjectCollection):
        for url in self.url_generator.generate_urls(webscraperObjectCollection):
            soup = fetch_soup_from_page(url)
            for parser in self.parsers:
                data = parser.parse_page(soup, {}, webscraperObjectCollection)
                for data_dict in data:
                    webscraperObjectCollection.databaseObject.tables[self.object_name].append(data_dict)


class WebscraperObjectFactory:

    def __init__(self, webscraper_object_dict, custom_objects):
        self.create_from_page_parser = None
        if 'create_from_page_parser' in webscraper_object_dict.keys():
            self.create_from_page_parser =  CreateFromPageParserFactory(webscraper_object_dict['create_from_page_parser']).create_from_page_parser
        if 'object_type' not in webscraper_object_dict.keys():
            parsers = []
            for parser_dict in webscraper_object_dict['parsers']:
                parsers.append(ParserObjectFactory(parser_dict).parser)
            url_generator = URLGeneratorFactory(webscraper_object_dict['urls'])
            self.webscraper = NewWebscraperObject(webscraper_object_dict['object_name'], parsers, url_generator.get_url_generator(), self.create_from_page_parser)
        elif webscraper_object_dict['object_type'] == 'custom_object':
            for custom_object in custom_objects:
                if custom_object.object_name == webscraper_object_dict['object_name']:
                    self.webscraper = custom_object
        else:
            raise Exception("No match for object type")


import unittest
import sys
sys.path.insert(0, '/Users/bradleystevenson/Programs/webscraper/src/bradleystevenson2015_webscraper')

from webscraper_object import WebscraperObjectCollection
from url_generator import URLGeneratorFactory

class TestDatabase(unittest.TestCase):

    def test_initialization(self):
        webscraperObjects = WebscraperObjectCollection("webscraper_schema.json", "../../databases/webscraper-database.db", "database_schema.json", [])
        webscraperObjects.run(['main.py'])

    def test_url_generator_base_only(self):
        url_generator = URLGeneratorFactory({ "base_url": "https://www.basketball-reference.com"}).get_url_generator()
        self.assertEqual(url_generator.generate_urls(None), ["https://www.basketball-reference.com"])

    def test_url_generator_hardcoded(self):
        url_generator = URLGeneratorFactory({"base_url": "https://www.basketball-reference.com/players/{replace}/", 'iterator': {'hardcoded': ['a', 'b']}}).get_url_generator()
        self.assertEqual(url_generator.generate_urls(None), ['https://www.basketball-reference.com/players/a/', 'https://www.basketball-reference.com/players/b/'])
if __name__ == '__main__':
    unittest.main()


from .common_webscraper_functions import get_element, get_children_element, row_has_link

class HTMLObjectIteratorFactory:


    def _get_narrow_down_function(self, function_name):
        if function_name == 'row_has_link':
            return row_has_link
        elif function_name == 'true':
            return true
        else:
            raise Exception("No match for narrow down function")
        
    def __init__(self, data_dict):
        self.data_dict = data_dict

    def create(self):
        base_object_fetcher = ObjectFetcherFactory(self.data_dict['base_object']).get_object_fetcher()
        children_object_fetcher = ChildrenObjectFetcher(self.data_dict['children_objects'])
        narrow_down_function = self._get_narrow_down_function(self.data_dict['narrow_down_function'])
        return HTMLObjectIterator(base_object_fetcher, children_object_fetcher, narrow_down_function)


class HTMLObjectIterator:

    def __init__(self, base_object_fetcher, children_object_fetcher, narrow_down_function):
        self.base_object_fetcher = base_object_fetcher
        self.children_object_fetcher = children_object_fetcher 
        self.narrow_down_function = narrow_down_function 

    def get_valid_elements(self, soup):
        return_array = []
        base_object = self.base_object_fetcher.fetch(soup)
        for eligible_element in self.children_object_fetcher.fetch_children(base_object):
            if self.narrow_down_function(eligible_element):
                return_array.append(eligible_element)
        return return_array




class ObjectFetcherFactory:

    def __init__(self, definition_dict):
        self.definition_dict = definition_dict

    def get_object_fetcher(self):
        child_fetcher = None
        if 'child' in self.definition_dict.keys():
            child_fetcher = ObjectFetcherFactory(self.definition_dict['child']).get_object_fetcher()
        return ObjectFetcher(self.definition_dict, child_fetcher)

class ObjectFetcher:

    def __init__(self, definition_dict, child_fetcher=None):
        self.element_function = get_element(definition_dict)
        self.child_fetcher = child_fetcher 


    def fetch(self, soup): 
        return_element = self.element_function(soup)
        if self.child_fetcher is not None:
            return_element =  self.child_fetcher.fetch(return_element)  
        return return_element

class ChildrenObjectFetcher:

    def __init__(self, definition_dict):
        self.children_element_function = get_children_element(definition_dict)

    def fetch_children(self, soup):
        return self.children_element_function(soup)